
// Note: Most of the code comes from the MacResearch OpenCL podcast

#include <CL/cl_ext.h>
#include "scl.hpp"

extern "C"
{
#include "png_util.h"
}

int main(int argc, const char * argv[]) {

  // define thread grid (only to be in line with CUDA version) 
  int Bre = 100;
  int Bim = 100;
  int Tre = 16;
  int Tim = 16;

  // number of pixels in real and imaginary directions 
  int Nre = Bre*Tre;
  int Nim = Bim*Tim;

  // parameters for mandelbrot
  const float centRe = -.7435669, centIm = .1314023;
  const float diam =.0022878;

  const float remin = centRe-0.5*diam;
  const float remax = centRe+0.5*diam;
  const float immin = centIm-0.5*diam;
  const float immax = centIm+0.5*diam;

  const float dre = (remax-remin)/(Nre-1.f);
  const float dim = (immax-immin)/(Nim-1.f);

  // HOST data
  float *h_image = new float[Nre*Nim];

  // initialize cl object on device devnumber
  int devnumber = 0;
  sclInfo scl(devnumber);

  // create kernel program  (function in .cl must match file name)
  sclKernel &ker  = scl.Kernel("mandelbrot.cl");

  // design kernel parameters and set up kernel function 
  sclFunction &mandelbrot = scl.Function(ker, sclNDRange(Nre, Nim), sclNDRange(1, 1));

  // create output buffer on DEVICE
  sclBuffer &d_image = scl.Buffer(sizeof(float)*Nre*Nim, h_image);  

  // execute kernel
  sclEvent event  = mandelbrot(remin, immin, dre, dim, d_image);

  // time event
  double elapsedTimeInSeconds = scl.timer(event);
  std::cout << "kernel took " << elapsedTimeInSeconds << std::endl;

  // copy data from device to host
  scl.readBuffer(d_image, h_image, Nre*Nim*sizeof(float));

  // Now write the file
  FILE *png = fopen("mandelbrotCL.png", "w");
  write_hot_png(png, Nre, Nim, h_image, 0, sqrt(160.));
  fclose(png);

  return 0;
}
