#ifndef __SCL
#define __SCL

// Note: Most of the code comes from the MacResearch OpenCL podcast

#define __CL_ENABLE_EXCEPTIONS

#include <cstring>
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>

using namespace std;

#include <cl.hpp>

#define sclDevice   cl::Device
#define sclPlatform cl::Platform
#define sclContext  cl::Context
#define sclQueue    cl::CommandQueue
#define sclProgram  cl::Program
#define sclBuffer   cl::Buffer
#define sclKernel   cl::Kernel
#define sclFunction cl::KernelFunctor
#define sclEvent    cl::Event
#define sclSource   cl::Program::Sources
#define sclNDRange  cl::NDRange
#define sclError    cl::Error

class sclInfo{
  
public:

  double start_t, end_t;

  std::vector<sclPlatform> platforms;
  std::vector<sclDevice>   devices;
  
  sclContext context;
  sclQueue  queue;

  sclQueue &getQueue(){
    return queue;
  }

  sclContext &getContext(){
    return context;
  }

  sclInfo(int devnumber){

    try {
      sclPlatform::get(&platforms);
      
      cl_context_properties cprops[3] = {CL_CONTEXT_PLATFORM, (cl_context_properties)(platforms[0])(), 0};

      context = sclContext(CL_DEVICE_TYPE_GPU, cprops, NULL, NULL);
      
      devices = context.getInfo<CL_CONTEXT_DEVICES>();

      queue = sclQueue(context, devices[devnumber], CL_QUEUE_PROFILING_ENABLE);

    }catch(sclError err){
      cerr << "ERROR: " << err.what() << "(" << err.err() << ")" << endl;
    }
    catch(string msg){
      cerr << "Exception caught in main(): " << msg << endl;
    }
  }
      
  sclKernel &Kernel(const char *filename){

    sclKernel *kernels = new sclKernel[1];
    sclProgram program;

    try{
      std::ifstream file(filename);
      std::string prog(std::istreambuf_iterator<char>(file),(std::istreambuf_iterator<char>()));

      sclProgram::Sources source = 
	sclProgram::Sources(1,std::make_pair(prog.c_str(), prog.length()+1));
      
      program = sclProgram(context, source);
      
      program.build(devices);
      
      char *name = strdup(filename);
      strtok(name, ".");
      kernels[0] = sclKernel(program, name);
      free(name);

    }
    catch(sclError err){

      cerr << "ERROR: " << err.what() << "(" << err.err() << ")" <<
	endl;
      if (err.err() == CL_BUILD_PROGRAM_FAILURE)
	{
	  string str =
	    program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(devices[0]);
	  cout << "Program Info: " << str << endl;
	}
    }
    catch(string msg){
      cerr << "Exception caught in main(): " << msg << endl;
    }

    return kernels[0];

  }

  sclFunction &Function(sclKernel &kernel, const sclNDRange &dimGrid, const sclNDRange &dimBlock){

    sclFunction *functors = new sclFunction[1];

    try{
      functors[0] = kernel.bind(queue, dimGrid, dimBlock);
    }
    catch(sclError err){
      cerr << "sclInfo::Function ERROR: " << err.what() << "(" << err.err() << ")" << endl;
    }
    catch(string msg){
      cerr << "sclInfo::Function Exception caught in main(): " << msg << endl;
    }
      
    return functors[0];
  }

  sclBuffer &Buffer(int sz, void *h_array){
    sclBuffer *bufs = new sclBuffer[1];
    
    bufs[0] = sclBuffer(context, CL_MEM_WRITE_ONLY | CL_MEM_USE_HOST_PTR, sz, h_array);

    return bufs[0];
  }

  void readBuffer(sclBuffer &d_array, void *h_array, int sz){
    
    try{
      queue.enqueueReadBuffer(d_array,CL_TRUE,0, sz, h_array);
    }
    catch(sclError err){
      cerr << "ERROR: " << err.what() << "(" << err.err() << ")" <<
	endl;
    }
    catch(string msg){
      cerr << "Exception caught in blockedDeviceToHost(): " << msg << endl;
    }
  }
  
  double timer(const sclEvent &event){

    // clear queue
    queue.finish();

    cl_ulong end   = event.getProfilingInfo <CL_PROFILING_COMMAND_END> ();
    cl_ulong start = event.getProfilingInfo <CL_PROFILING_COMMAND_START> ();
    return (end - start) * 1.0e-9;
  }

};

#endif
